@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use base qw( Exporter );
 
-our $VERSION = '0.09_02';
+our $VERSION = '0.15_01';
 
 use AnyEvent::Stomper::Frame;
 use AnyEvent::Stomper::Error;
@@ -67,19 +67,27 @@ sub new {
 
   my $self = bless {}, $class;
 
-  $self->{host} = $params{host} || D_HOST;
-  $self->{port} = $params{port} || D_PORT;
-  $self->{login}    = $params{login};
-  $self->{passcode} = $params{passcode};
-  $self->{vhost}    = $params{vhost};
+  $self->{host}            = $params{host} || D_HOST;
+  $self->{port}            = $params{port} || D_PORT;
+  $self->{login}           = $params{login};
+  $self->{passcode}        = $params{passcode};
+  $self->{vhost}           = $params{vhost};
+  $self->{lazy}            = $params{lazy};
+  $self->{handle_params}   = $params{handle_params} || {};
+  $self->{default_headers} = $params{default_headers} || {};
+  $self->{body_encoder}    = $params{body_encoder};
+  $self->{body_decoder}    = $params{body_decoder};
+  $self->{on_connect}      = $params{on_connect};
+  $self->{on_disconnect}   = $params{on_disconnect};
 
   if ( defined $params{heartbeat} ) {
     unless ( ref( $params{heartbeat} ) eq 'ARRAY' ) {
-      croak qq{"heartbeat" must be specified as array reference};
+      croak q{"heartbeat" must be specified as array reference};
     }
+
     foreach my $val ( @{ $params{heartbeat} } ) {
       if ( $val =~ /\D/ ) {
-        croak qq{"heartbeat" values must be an integer numbers};
+        croak q{"heartbeat" values must be an integer numbers};
       }
     }
 
@@ -89,10 +97,17 @@ sub new {
     $self->{heartbeat} = D_HEARTBEAT;
   }
 
-  $self->{lazy}          = $params{lazy};
-  $self->{handle_params} = $params{handle_params} || {};
-  $self->{on_connect}    = $params{on_connect};
-  $self->{on_disconnect} = $params{on_disconnect};
+  if ( defined $params{command_headers} ) {
+    unless ( ref( $params{command_headers} ) eq 'HASH' ) {
+      croak q{"command_headers" must be specified as hash reference};
+    }
+
+    my %command_headers;
+    while ( my ( $cmd_name, $headers ) = each %{ $params{command_headers} } ) {
+      $command_headers{ uc($cmd_name) } = $headers;
+    }
+    $self->{command_headers} = \%command_headers;
+  }
 
   $self->connection_timeout( $params{connection_timeout} );
   $self->reconnect_interval( $params{reconnect_interval} );
@@ -121,14 +136,6 @@ sub execute {
   return;
 }
 
-sub force_disconnect {
-  my $self = shift;
-
-  $self->_disconnect();
-
-  return;
-}
-
 # Generate methods
 {
   no strict qw( refs );
@@ -147,6 +154,14 @@ sub force_disconnect {
   }
 }
 
+sub force_disconnect {
+  my $self = shift;
+
+  $self->_disconnect();
+
+  return;
+}
+
 sub on_error {
   my $self = shift;
 
@@ -337,9 +352,9 @@ sub _create_on_read {
       return if $handle->destroyed;
 
       if ( defined $cmd_name ) {
-        my $content_length = $headers->{'content-length'};
-
-        if ( defined $content_length ) {
+        my $content_length;
+        if ( defined $headers->{'content-length'} ) {
+          $content_length = $headers->{'content-length'};
           return if length( $handle->{rbuf} ) < $content_length + 1;
         }
         else {
@@ -348,6 +363,18 @@ sub _create_on_read {
         }
 
         my $body = substr( $handle->{rbuf}, 0, $content_length, '' );
+        if ( defined $self->{body_decoder}
+          && length($body) > 0 )
+        {
+          $body = $self->{body_decoder}->(
+            $body,
+
+            defined $headers->{'content-type'}
+            ? $headers->{'content-type'}
+            : ()
+          );
+        }
+
         $handle->{rbuf} =~ s/^\0(?:${\(RE_EOL)})*//;
 
         $frame = _new_frame( $cmd_name, $headers, $body );
@@ -361,6 +388,7 @@ sub _create_on_read {
         return unless $handle->{rbuf} =~ s/^(.+?)(?:${\(RE_EOL)}){2}//s;
 
         ( $cmd_name, my @header_strings ) = split( m/${\(RE_EOL)}/, $1 );
+
         foreach my $header_str (@header_strings) {
           my ( $name, $value ) = split( /:/, $header_str, 2 );
           $headers->{ _unescape($name) } = _unescape($value);
@@ -390,18 +418,28 @@ sub _prepare {
       $cbs{on_receipt} = pop @{$args};
     }
   }
-  my %cmd_headers = @{$args};
-  if ( defined $cmd_headers{on_receipt} ) {
-    $cbs{on_receipt} = delete $cmd_headers{on_receipt};
+
+  my %headers = @{$args};
+  foreach my $name ( qw( on_receipt on_message ) ) {
+    if ( defined $headers{$name} ) {
+      $cbs{$name} = delete $headers{$name};
+    }
   }
-  if ( defined $cmd_headers{on_message} ) {
-    $cbs{on_message} = delete $cmd_headers{on_message};
-  }
-  my $body = delete $cmd_headers{body};
+  my $body = delete $headers{body};
+
+  %headers = (
+    %{ $self->{default_headers} },
+
+    defined $self->{command_headers}{$cmd_name}
+    ? %{ $self->{command_headers}{$cmd_name} }
+    : (),
+
+    %headers,
+  );
 
   my $cmd = {
     name    => $cmd_name,
-    headers => \%cmd_headers,
+    headers => \%headers,
     body    => $body,
     %cbs,
   };
@@ -414,7 +452,7 @@ sub _prepare {
       my $err     = shift;
 
       if ( defined $err ) {
-        $self->{on_error}->( $err, $receipt );
+        $self->{on_error}->($err);
         return;
       }
     };
@@ -450,6 +488,8 @@ sub _execute {
         && $self->{reconnect_interval} > 0 )
       {
         unless ( defined $self->{_reconnect_timer} ) {
+          weaken($self);
+
           $self->{_reconnect_timer} = AE::timer(
             $self->{reconnect_interval}, 0,
             sub {
@@ -478,39 +518,54 @@ sub _push_write {
   my $self = shift;
   my $cmd  = shift;
 
-  my $cmd_headers = $cmd->{headers};
+  my $headers = $cmd->{headers};
 
   my $need_receipt;
   if ( exists $NEED_RECEIPT{ $cmd->{name} }
-    || defined $cmd_headers->{receipt} )
+    || defined $headers->{receipt} )
   {
     $need_receipt = 1;
     if ( $cmd->{name} eq 'CONNECT' ) {
       $self->{_pending_receipts}{CONNECTED} = $cmd;
     }
     else {
-      if ( !defined $cmd_headers->{receipt}
-        || $cmd_headers->{receipt} eq 'auto' )
+      if ( !defined $headers->{receipt}
+        || $headers->{receipt} eq 'auto' )
       {
-        $cmd_headers->{receipt} = $self->{_session_id} . '@@'
+        $headers->{receipt} = $self->{_session_id} . '@@'
             . $self->{_receipt_seq}++;
       }
-      $self->{_pending_receipts}{ $cmd_headers->{receipt} } = $cmd;
+      $self->{_pending_receipts}{ $headers->{receipt} } = $cmd;
     }
   }
 
-  unless ( defined $cmd->{body} ) {
-    $cmd->{body} = '';
+  my $body = $cmd->{body};
+  unless ( defined $body ) {
+    $body = '';
   }
-  unless ( defined $cmd_headers->{'content-length'} ) {
-    $cmd_headers->{'content-length'} = length( $cmd->{body} );
+  if ( defined $self->{body_encoder}
+    && length($body) > 0 )
+  {
+    $body = $self->{body_encoder}->(
+      $body,
+
+      defined $headers->{'content-type'}
+      ? $headers->{'content-type'}
+      : ()
+    );
+  }
+  unless ( defined $headers->{'content-length'} ) {
+    $headers->{'content-length'} = length($body);
   }
 
-  my $frame_str = uc( $cmd->{name} ) . EOL;
-  while ( my ( $name, $value ) = each %{$cmd_headers} ) {
+  my $frame_str = $cmd->{name} . EOL;
+  while ( my ( $name, $value ) = each %{$headers} ) {
+    unless ( defined $value ) {
+      $value = '';
+    }
     $frame_str .= _escape($name) . ':' . _escape($value) . EOL;
   }
-  $frame_str .= EOL . $cmd->{body} . "\0";
+  $frame_str .= EOL . "$body\0";
 
   $self->{_handle}->push_write($frame_str);
 
@@ -686,8 +741,8 @@ sub _process_receipt {
   }
 
   if ( exists $SUBUNSUB_CMDS{ $cmd->{name} } ) {
-    my $cmd_headers = $cmd->{headers};
-    my $sub_id = $cmd_headers->{id} || $cmd_headers->{destination};
+    my $headers = $cmd->{headers};
+    my $sub_id  = $headers->{id} || $headers->{destination};
 
     if ( $cmd->{name} eq 'SUBSCRIBE' ) {
       $self->{_subs}{$sub_id} = $cmd;
@@ -1004,8 +1059,16 @@ Disabled by default.
 
 =item reconnect_interval => $reconnect_interval
 
-If the parameter is specified, the client will try to reconnect only after
-this interval. Commands executed between reconnections will be queued.
+If the connection to the server was lost, the client will try to restore the
+connection when you execute next command. By default reconnection is performed
+immediately, on next command execution. If the C<reconnect_interval> parameter
+is specified, the client will try to reconnect only after this interval and
+commands executed between reconnections will be queued.
+
+The client will try to reconnect only once and, if attempt fails, the error
+object is passed to command callback. If you need several attempts of the
+reconnection, you must retry a command from the callback as many times, as you
+need.
 
   reconnect_interval => 5,
 
@@ -1020,9 +1083,53 @@ Specifies L<AnyEvent::Handle> parameters.
     linger   => 60,
   }
 
-Enabling of the C<autocork> parameter can improve perfomance. See
+Enabling of the C<autocork> parameter can improve performance. See
 documentation on L<AnyEvent::Handle> for more information.
 
+=item default_headers => \%headers
+
+Specifies default headers for all outgoing frames.
+
+  default_headers => {
+    'x-foo' => 'foo_value',
+    'x-bar' => 'bar_value',
+  }
+
+=item command_headers
+
+Specifies default headers for particular commands.
+
+  command_headers => {
+    SEND => {
+      receipt => 'auto',
+    },
+
+    SUBSCRIBE => {
+      durable => 'true',
+      ack     => 'client',
+    },
+  }
+
+=item body_encoder => $cb->( $body [, $content_type ] )
+
+Specifies the encode function for the body of the frame. The function accepts
+two arguments: the body and the content type of the body if it specified in the
+frame. The function must return the encoded body.
+
+  body_encoder => sub {
+    return encode_json( $_[0] );
+  }
+
+=item body_decoder => $cb->( $body [, $content_type ] )
+
+Specifies the decode function for the body of the frame. The function accepts
+two arguments: the body and the content type of the body if it specified in the
+frame. The function must return the decoded body.
+
+  body_decoder => sub {
+    return decode_json( $_[0] );
+  }
+
 =item on_connect => $cb->()
 
 The C<on_connect> callback is called when the connection is successfully
@@ -1049,25 +1156,28 @@ an error messages to C<STDERR>.
 
 =head1 COMMAND METHODS
 
-To execute the STOMP command you must call appropriate method. The client
-automatically adds C<content-length> header to all outgoing frames. The body of
-the frame you can specify in C<body> parameter of the command method.
+To execute the STOMP command you must call appropriate method. STOMP headers
+can be specified as command parameters. The client automatically adds
+C<content-length> header to all outgoing frames. Every command method can also
+accept two additional parameters: the C<body> parameter where you can specify
+the body of the frame, and the C<on_receipt> parameter that is the alternative
+way to specify the command callback.
 
 If you want to receive C<RECEIPT> frame, you must specify C<receipt> header.
-The C<receipt> header can take special value C<auto>. In this case the
-value for C<receipt> header will be generated automatically by the client.
+The C<receipt> header can take the special value C<auto>. If it set, the
+receipt identificator will be generated automatically by the client. The
 C<RECEIPT> frame is passed to the command callback in first argument as the
-object of the class L<AnyEvent::Stomper::Frame>. If the C<receipt> header
-is not specified the first argument of the command callback will be C<undef>.
+object of the class L<AnyEvent::Stomper::Frame>. If the C<receipt> header is
+not specified the first argument of the command callback will be C<undef>.
 
 For commands C<SUBSCRIBE>, C<UNSUBSCRIBE>, C<DISCONNECT> the client
 automatically adds C<receipt> header for internal usage.
 
-The command callback is called after successful sending of the command to the
-server or when C<RECEIPT> frame will be received if C<receipt> header
-was specified. If any error occurred during the command execution, the error
-object is passed to the callback in second argument. Error object is the
-instance of the class L<AnyEvent::Stomper::Error>.
+The command callback is called in one of two cases depending on the presence of
+the C<receipt> header: when the command was successfully sent to the server or
+when the C<RECEIPT> frame will be received. If any error occurred during the
+command execution, the error object is passed to the callback in second
+argument. Error object is the instance of the class L<AnyEvent::Stomper::Error>.
 
 The command callback is optional. If it is not specified and any error
 occurred, the C<on_error> callback of the client is called.
@@ -1076,7 +1186,7 @@ The full list of all available headers for every command you can find in STOMP
 protocol specification and in documentation on your STOMP server. For various
 versions of STOMP protocol and various STOMP servers they can be differ.
 
-=head2 send( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 send( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 Sends a message to a destination in the messaging system.
 
@@ -1127,7 +1237,7 @@ Sends a message to a destination in the messaging system.
     }
   );
 
-=head2 subscribe( [ %headers ] [, %params ] [, $cb->( $msg ) ] )
+=head2 subscribe( [ %params ] [, $cb->( $msg ) ] )
 
 The method is used to register to listen to a given destination. The
 C<subscribe> method require the C<on_message> callback, which is called on
@@ -1180,7 +1290,7 @@ callback, this callback will be act as C<on_message> callback.
     }
   );
 
-=head2 unsubscribe( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 unsubscribe( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 The method is used to remove an existing subscription.
 
@@ -1204,7 +1314,7 @@ The method is used to remove an existing subscription.
     }
   );
 
-=head2 ack( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 ack( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 The method is used to acknowledge consumption of a message from a subscription
 using C<client> or C<client-individual> acknowledgment. Any messages received
@@ -1233,7 +1343,7 @@ message has been acknowledged via an C<ack()> method.
     }
   );
 
-=head2 nack( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 nack( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 The C<nack> method is the opposite of C<ack> method. It is used to tell the
 server that the client did not consume the message.
@@ -1260,19 +1370,19 @@ server that the client did not consume the message.
     }
   );
 
-=head2 begin( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 begin( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 The method C<begin> is used to start a transaction.
 
-=head2 commit( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 commit( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 The method C<commit> is used to commit a transaction.
 
-=head2 abort([ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 abort([ %params ] [, $cb->( $receipt, $err ) ] )
 
 The method C<abort> is used to roll back a transaction.
 
-=head2 disconnect( [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 disconnect( [ %params ] [, $cb->( $receipt, $err ) ] )
 
 A client can disconnect from the server at anytime by closing the socket but
 there is no guarantee that the previously sent frames have been received by
@@ -1280,7 +1390,7 @@ the server. To do a graceful shutdown, where the client is assured that all
 previous frames have been received by the server, you must call C<disconnect>
 method and wait for the C<RECEIPT> frame.
 
-=head2 execute( $command, [ %headers ] [, %params ] [, $cb->( $receipt, $err ) ] )
+=head2 execute( $command, [ %params ] [, $cb->( $receipt, $err ) ] )
 
 An alternative method to execute commands. In some cases it can be more
 convenient.
@@ -1387,14 +1497,14 @@ Gets or sets the C<on_error> callback.
 The method for forced disconnection. All uncompleted operations will be
 aborted.
 
-=head1 CONNECTION POOL
+=head1 WORKING WITH CLUSTER
 
-If you have the cluster or set of STOMP servers, you can use
-L<AnyEvent::Stomper::Pool> to work with them.
+If you have the cluster of STOMP servers, you can use
+L<AnyEvent::Stomper::Cluster> to work with it.
 
 =head1 SEE ALSO
 
-L<AnyEvent::Stomper::Pool>
+L<AnyEvent::Stomper::Cluster>
 
 =head1 AUTHOR
 
